/* File:      corr.C
   Date:      3/21/02  (written fall 2001)
   Author:    C. Paciorek
   Purpose:   calculates various correlation functions as function of Euclidean or angular distance
   Compile:   g++ -c
   Usage:     
   Arguments: 
   Requires:  base.h corr.h
   Reference: Das 2000 U Wash Stats thesis (eqn 3.32, p.37); Gneiting 2001 (J Multivar Anal in press); Handcock & Stein Technometrics 1993 (for Matern)
*/

#include "base.h"
#include "corr.h"

double gneiting_corr_ang(double ang_distance,double max_ang_distance){
  // calcs compactly supported Gneiting correlation function
  // if requested, the 2*sin business converts ang distance to Euclidean distance in R^3 and the division puts the value in (0,1), where locations->range is the max distance for which corr is not zero
  double tmp=TWO_PI*sin(ang_distance/2.0)/sin(max_ang_distance/2.0);
  return((1-tmp/TWO_PI)*sin(tmp)/tmp+(1-cos(tmp))/(PI*tmp));
}

double das_corr(double ang_distance, double parm){
  // calcs Das' spherical correlation function - equ. 3.32 in Das 2000, but I have calculated a correlation function from her covariance function
  return(   ( pow(1-parm,3.0))  /  ( pow(1-2*parm*cos(ang_distance)+parm*parm,1.5)));
}

double matern_corr_euc(double euc_distance,double theta1,double theta2){
  // theta2 is nu the smoothness parameter
  // 2*sqrt(theta2) is so that as theta2->\infty, correlation fxn approaches squ exponential; with theta2=1/2; this is equivalent to exp(-2*sqrt(theta2)*dist/theta1)
  double dx=2.0*sqrt(theta2)*euc_distance/theta1;

  if(dx>500){   // rkbesl underflows for values larger than about 700
    return(0.0);
  }
  if((dx<1e-8 && theta2>=15.0) || (dx<1e-20 && theta2<15.0)){ // rkbesl overflow depends on theta2
    return(1.0); 
  }

  int nb=1+(int)floor(theta2);
  double alpha=(double)theta2-nb+1;
  int ize=1;
  int ncalc;
  double* tmp=new double[nb];
  rkbesld_(&dx,&alpha,&nb,&ize,tmp,&ncalc);
  // end for rkbesld
  double outrkb=tmp[nb-1];
  delete [] tmp;

 double outval=exp(-lgamma(theta2)-(theta2-1.0)*log(2.0)+theta2*log(dx)+log(outrkb));  // use out2 in place of tmp[nb-1] for bessikD
  if(outval>1.0){
    cerr << setprecision(16) << "Warning; correlation exceeds 1.0 " << outval << " " << dx << " " << theta2 << " " << euc_distance << endl;
  }

  return(outval);  
}
/* old bessikD version
  bessikD(dx,theta2,&out1,&out2,&out3,&out4);
  return(exp(-lgamma(theta2)-(theta2-1.0)*log(2.0)+theta2*log(dx)+log(out2)));
*/

double squexp_corr_ang(double ang_distance,double parm){
  double dist=sin(ang_distance/2.0);
  return(exp(-dist*dist/(2.0*parm*parm)));
}

double matern_corr_ang(double ang_distance,double theta1,double theta2){
  // theta2 is nu the smoothness parameter
  // for fixed theta2 I can pre-calculate lgamma(theta2)+(theta2-1)*log(2)
  // Double precision version
  double dx = 2.0*sqrt(theta2)*sin(ang_distance/2.0)/theta1;
  
  if(dx>500){   // rkbesl underflows for values larger than about 700
    return(0.0);
  }
  if((dx<1e-8 && theta2>=15.0) || (dx<1e-20 && theta2<15.0)){ // rkbesl overflow depends on theta2
    return(1.0); 
  }
  // results get pretty flaky for small dx, so not sure I would trust even the cutoffs above, hence the line below checking for corr > 1.0

  int nb=1+(int)floor(theta2);
  double alpha=(double)theta2-nb+1;
  int ize=1;
  int ncalc;
  double* tmp=new double[nb];
  rkbesld_(&dx,&alpha,&nb,&ize,tmp,&ncalc);
  // end for rkbesld
  double outrkb=tmp[nb-1];
  delete [] tmp;

 double outval=exp(-lgamma(theta2)-(theta2-1.0)*log(2.0)+theta2*log(dx)+log(outrkb));  // use out2 in place of tmp[nb-1] for bessikD
  if(outval>1.0){
    cerr << setprecision(16) << "Warning; correlation exceeds 1.0 " << outval << " " << dx << " " << theta2 << " " << ang_distance << endl;
    outval=1.0;
  }
  return(outval);  
}

  /* code is old code for bessikD version; regr model showed that bessikD is not accurate beyond 7 or 8 digits for some reason based on not being able to reconstruct C=L*t(L)
  double out1,out2,out3,out4;
  double dx = 2.0*sqrt(theta2)*sin(ang_distance/2.0)/theta1;
  // old version of Matern form
  //double dx =  sin(ang_distance/2.0)/(2.0*theta1);
  bessikD(dx,theta2,&out1,&out2,&out3,&out4);
  return(exp(-lgamma(theta2)-(theta2-1.0)*log(2.0)+theta2*log(dx)+log(out2)));
  */

/* code below is for discretized, float (single precision), and pre-computing some values */
/*
  // Float with unknown theta2
  if(0==1){
  if(!theta1 && !theta2){
    return(0);
  }
  else{
  // theta2 not fixed
  int nb,ize,ncalc;
  float alpha,x;
  double dx,output;
  nb=1+(int)floor(theta2);
  alpha=(float)theta2-nb+1;
  dx=sin(ang_distance/2)/(2.0*theta1);
  x=(float)dx;
  ize=1;
  float* tmp=new float[nb];
  rkbesl_(&x,&alpha,&nb,&ize,tmp,&ncalc);
  output=exp(-lgamma(theta2)-(theta2-1.0)*log(2.0)+theta2*log(dx)+log((double)tmp[nb-1]));
  return(output);
  }
  }
*/
/*
  // Float with known theta2
  if(!theta1 && !theta2){
    return(0);
  }
  else{
  int nb=3,ize,ncalc;
  float alpha,x;
  double dx,output;
  //nb=1+(int)floor(theta2);
  //alpha=(float)theta2-nb+1;
  alpha=0.0;
  dx=sin(ang_distance/2.0)/(2.0*theta1);
  x=(float)dx;
  ize=1;
  float tmp[3];
  rkbesl_(&x,&alpha,&nb,&ize,tmp,&ncalc);
  output=exp(-lgamma(theta2)-(theta2-1.0)*log(2.0)+theta2*log(dx)+log((double)tmp[nb-1]));
  return(output);
  }
  */
  // Discretized float
  /*
  int nb,ize,ncalc,whichval;
  float alpha,x;
  double dx,output;
  if(!theta1 && !theta2){
    theta2=2.0;
    nb=3;
    alpha=0.0;
    ize=1;
    float* tmp=new float[nb];
    for(int i=1;i<25000000;i++){
      x=(float)i/1000000.0;
      rkbesl_(&x,&alpha,&nb,&ize,tmp,&ncalc);
      besselvec[i]=exp(-lgamma(theta2)-(theta2-1.0)*log(2.0)+theta2*log(x)+log((double)tmp[2]));
      output=0.0;
    }
    besselvec[0]=1.0;
  }
  else{
    nb=3;
    alpha=0.0;
    dx=sin(ang_distance/2)/(2.0*theta1);
    x=(float)dx;
    ize=1;
    whichval=(int)ceil(dx*1000000.0);
    if(whichval-dx*1000000.0>0.5){
      whichval--;
    }
    if(dx>=25.0){
      whichval=24999999;
    }
    if(dx<0.000001){
      whichval=0;
    }
    output=besselvec[whichval];
  }
  
  return(output);
  */







